#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <malloc.h>
#include <string.h>
#include <ctype.h>

#include "globaldefn.h"
#include "stream.h"

#include <kqml.h>

int C_kqmlInit FN_PROTOTYPE((int ans_flag, char *name, SP_term *SP_stream, 
	char **agent_name, char **ans_name, char **protocol, char **addr_info));
int C_kqmlRegister FN_PROTOTYPE((char *performative, int blocking));
int C_kqmlDeliver FN_PROTOTYPE((SP_term SP_msg));
int C_kqmlSend FN_PROTOTYPE((int flag, int timeout_value, SP_term SP_msg, 
	SP_term *SP_reply));
int C_kqmlParse FN_PROTOTYPE((int len, SP_term SP_list, SP_term *SP_msg,
	SP_term *SP_string));
int C_kqmlSprintMsg FN_PROTOTYPE((SP_term SP_list, SP_term *SP_string));
void C_kqmlDebug FN_PROTOTYPE((int level, char *message));

static char *strlwr (char *string);

static void (*OldSigHandler) FN_PROTOTYPE((int sigNo));
static void NewSigHandler FN_PROTOTYPE((int sigNo));
static int SpecialHandler FN_PROTOTYPE((int sigNo));

static void *generic_handler (char *content, kqml_message *msg,
	KQML_REPLY_TYPE *reply_type);

static void string_to_list (int len, char *string, SP_term *SP_list);
static char *list_to_string (int len, SP_term SP_list);
static void kmsg_to_list (kqml_message *msg, SP_term *SP_list);
static kqml_message *list_to_kmsg (SP_term SP_list);
static void *construct_reply_list (SP_term SP_reply_list, 
	KQML_REPLY_TYPE *reply_type);

static SP_pred_ref SP_handler = NULL, SP_ans_handler = NULL;
static ComsStream stream;

static int
C_ans_name_lookup (char *agent, char **name, char **protocol, 
	char **addr_info)
{
	static char *i_name = NULL, *i_protocol = NULL, *i_addr_info = NULL;
	SP_term SP_agent, SP_name_len, SP_name, SP_protocol_len, SP_protocol;
	SP_term SP_addr_info_len, SP_addr_info;
	long name_len, protocol_len, addr_info_len;
	int ret_value;

	if (i_name)
	{
		free (i_name);
		i_name = NULL;
	}

	if (i_protocol)
	{
		free (i_protocol);
		i_protocol = NULL;
	}

	if (i_addr_info)
	{
		free (i_addr_info);
		i_addr_info = NULL;
	}

	SP_show_term (&SP_agent);
	SP_show_term (&SP_name_len);
	SP_show_term (&SP_name);
	SP_show_term (&SP_protocol_len);
	SP_show_term (&SP_protocol);
	SP_show_term (&SP_addr_info_len);
	SP_show_term (&SP_addr_info);

	/* Prepare the args before calling prolog */
	string_to_list (strlen (agent), agent, &SP_agent);

	if ((ret_value = SP_query(SP_ans_handler, &SP_agent, &SP_name_len, &SP_name,
		&SP_protocol_len, &SP_protocol, &SP_addr_info_len, &SP_addr_info)) <= 0)
	{
		fprintf (stderr, "SP: ANS Handler failed with %d\n", ret_value);
		ret_value = FALSE;
	}
	else
	{
		ret_value = TRUE;
		if (SP_get_integer (SP_name_len, &name_len) == 0)
		{
			fprintf (stderr, "SP: ANS Handler: Get name len failed.\n");
			ret_value = FALSE;
		}
		else if (SP_get_integer (SP_protocol_len, &protocol_len) == 0)
		{
			fprintf (stderr, "SP: ANS Handler: Get protocol len failed.\n");
			ret_value = FALSE;
		}
		else if (SP_get_integer (SP_addr_info_len, &addr_info_len) == 0)
		{
			fprintf (stderr, "SP: ANS Handler: Get addr_info len failed.\n");
			ret_value = FALSE;
		}
		else 
		{
			i_name = *name = list_to_string (name_len, SP_name);
			if (*name == NULL)
			{
				fprintf (stderr, "SP: ANS Handler: Get name failed.\n");
				ret_value = FALSE;
			}
			i_protocol = *protocol = list_to_string (protocol_len, SP_protocol);
			if (*protocol == NULL)
			{
				fprintf (stderr, "SP: ANS Handler: Get protocol failed.\n");
				ret_value = FALSE;
			}

			i_addr_info = *addr_info = 
				list_to_string (addr_info_len, SP_addr_info);
			if (*addr_info == NULL)
			{
				fprintf (stderr, "SP: ANS Handler: Get addr_info failed.\n");
				ret_value = FALSE;
			}
		}
	}

	SP_hide_term (&SP_addr_info);
	SP_hide_term (&SP_addr_info_len);
	SP_hide_term (&SP_protocol);
	SP_hide_term (&SP_protocol_len);
	SP_hide_term (&SP_name);
	SP_hide_term (&SP_name_len);
	SP_hide_term (&SP_agent);

	return ret_value;
}

int
C_kqmlInit(int ans_flag, char *name, SP_term *SP_stream, char **agent_name,
	char **ans_name, char **protocol, char **addr_info)
{
	static char initDone = FALSE;
	ans_handler ans_lookup_func;

	if (initDone)
	{
		fprintf (stderr, "SP: Warning, init already done: [%s]\n", name);
		return FALSE;
	}

	/* Locate the default prolog handler */
	SP_handler = SP_predicate ("handler", 3, "pkril");
	if (SP_handler == NULL)
	{
		fprintf (stderr, "SP: Could not locate pkril:handler/3\n");
		return FALSE;
	}

	/* Create a stream for converting prolog term to strings and vice versa */
	if (Stream_open (&stream) == ERR)
	{
		fprintf (stderr, "SP: Could not open conversion stream\n");
		return FALSE;
	}
	*SP_stream = stream.SP_streamPtr;

	if (ans_flag)
	{
		ans_lookup_func = C_ans_name_lookup;
		SP_ans_handler = SP_predicate ("ans_handler", 7, "pkril");
		if (SP_ans_handler == NULL)
		{
			fprintf (stderr, "SP: Could not locate pkril:ans_handler/7\n");
			return FALSE;
		}
	}
	else
	{
		ans_lookup_func = KQML_NOT_AN_ANS;
	}

	*agent_name =
		*ans_name = 
		*protocol = 
		*addr_info = NULL;

	if (kqml_initialize(name, ans_lookup_func) == ERR)
	{
		fprintf (stderr, "SP: kqml_initialize failed\n");
		return FALSE;
	}

	*agent_name = kqml_agent_name;
	*ans_name = kqml_ans_name;
	*protocol = kqml_protocol;
	*addr_info = kqml_addr_info;

	initDone = TRUE;

	/* kludge 
		signal handlers should be installed through prolog, to let prolog
		know that we are using signal handlers.
	*/

	if ((OldSigHandler = signal(SIGIO, SIG_IGN)) == (void *)ERR)
	{
		fprintf (stderr, "SP: Could not install SIGIO wrapper\n");
		return FALSE;
	}

	/* Install the handler through prolog */
	SP_signal(SIGIO, NewSigHandler);

	return TRUE;
}

int
C_kqmlRegister(char *performative, int blocking)
{
	return kqml_register (generic_handler, performative, blocking);
}

void
C_kqmlDebug(int level, char *message)
{
	_kqml_dbg_msg (level)(message);
}

int
C_kqmlParse(int len, SP_term SP_list, SP_term *SP_msg, SP_term *SP_string)
{
	kqml_message *msg;
	int ret_value, used_len;
	char *string;

	if (len <= 0)
	{
		return FALSE;
	}

	string = list_to_string (len, SP_list);
	if (string == NULL)
	{
		fprintf (stderr, "SP: kqmlParse: String Conversion Error\n");
		return FALSE;
	}

	ret_value = kqml_parse_buffer (string, len, &used_len, &msg);
	string_to_list (len - used_len, string+used_len, SP_string);
	free (string);

	if (ret_value != KQML_PARSE_OK)
	{
		return FALSE;
	}
	else
	{
		kmsg_to_list (msg, SP_msg);
		kqml_free_msg (msg);
		return TRUE;
	}
}

int
C_kqmlSprintMsg(SP_term SP_list, SP_term *SP_string)
{
	kqml_message *msg;
	int len;
	char *string;

	msg = list_to_kmsg (SP_list);
	if (msg)
	{
		string = kqml_sprint_msg (NULL, &len, msg);
		if (string == NULL)
		{
			return FALSE;
		}
		string_to_list (len-1, string, SP_string);
		free (string);
		return TRUE;
	}
	else
		return FALSE;
}

int
C_kqmlSend(int blockingFlag, int timeout_value, SP_term SP_msg, 
	SP_term *SP_reply)
{
	kqml_message *msg, *reply;
	int ret_value;

	msg = list_to_kmsg (SP_msg);
	if (msg == NULL)
	{
		return -1;
	}

	if (blockingFlag)
	{
		ret_value = kqml_send (timeout_value, msg, &reply);
	}
	else
	{
		reply = NULL;
		ret_value = kqml_send (timeout_value, msg, NULL);
	}

	kqml_free_msg (msg);

	if (ret_value >= 0)
	{
		if (reply)
		{
			kmsg_to_list (reply, SP_reply);
			kqml_free_msg (reply);
		}
		else
		{
			SP_put_string (SP_reply, "[]");
		}
	}
	return ret_value;
}

int
C_kqmlDeliver(SP_term SP_msg)
{
	kqml_message *msg;
	int ret_value;

	msg = list_to_kmsg (SP_msg);
	if (msg == NULL)
	{
		return -1;
	}

	ret_value = kqml_deliver (msg);
	kqml_free_msg (msg);

	return ret_value;
}

static void *generic_handler (char *content, kqml_message *msg,
	KQML_REPLY_TYPE *reply_type)
{
	void *reply;
	SP_term SP_msg, SP_reply_list, SP_content;
	int ret_value, content_len;

	SP_show_term (&SP_content);
	SP_show_term (&SP_msg);
	SP_show_term (&SP_reply_list);

	content_len = 0;
	/* We just want to find out the length of the content */
	kqml_get_field (":content", msg, &content_len, NULL);

	/* Prepare the args before calling prolog */
	string_to_list (content_len, content, &SP_content);
	kmsg_to_list (msg, &SP_msg);
	SP_put_variable (&SP_reply_list);

	if ((ret_value = 
		SP_query(SP_handler, &SP_content, &SP_msg, &SP_reply_list)) <= 0)
	{
		fprintf (stderr, "SP: Handler failed with %d\n", ret_value);
		*reply_type = KQML_REPLY_ERROR;
		reply = NULL;
	}
	else
	{
		reply = construct_reply_list (SP_reply_list, reply_type);
	}

	SP_hide_term (&SP_content);
	SP_hide_term (&SP_msg);
	SP_hide_term (&SP_reply_list);

	return reply;
}

static void *
construct_reply_list (SP_term SP_reply_list, KQML_REPLY_TYPE *reply_type)
{
	int msg_no, arity;
	char *comp_name;
	SP_term SP_kmsg;
	kqml_message *kmsg;
	KQML_KMSG_LIST *reply_list;

	reply_list = NULL;
	/* Check if the prolog list is empty. If it is empty, it will
		be an atom and the name of the atom will be [] or nil
	*/
	if (SP_is_atom (SP_reply_list))
	{
		if (SP_get_string (SP_reply_list, &comp_name) == 0 ||
			(strcmp (comp_name, "[]") != 0 && strcmp (comp_name, "nil") != 0))
		{
			fprintf (stderr, "SP: Not a valid kqml list structure ?\n");
			*reply_type = KQML_REPLY_ERROR;
			return NULL;
		}
		else
		{
			*reply_type = KQML_REPLY_NONE;
			return NULL;
		}
	}
	
	/* If it is not empty, it should be a compound item of type
		.(kmsg, rest)
	*/
	msg_no = 0;
	while (++msg_no, TRUE)
	{
		/* If this item is the empty list, then we have reached the end 
			of the list */
		if (SP_is_atom (SP_reply_list))
		{
			if (SP_get_string (SP_reply_list, &comp_name) == 0 ||
				strcmp (comp_name, "[]") != 0)
			{
				fprintf (stderr, "SP: %d Not a valid kqml list structure ?\n", 
					msg_no);
				kqml_free_kmsg_list (reply_list);
				*reply_type = KQML_REPLY_NONE;
				return NULL;
			}
			else
			{
				*reply_type = KQML_REPLY_KMSG_LIST;
				break;
			}
		}

		/* If this item is not the empty list, it should be a 
			.(kmsg, res)
			
		*/
		if (! SP_is_compound (SP_reply_list) ||
			SP_get_compound (SP_reply_list, &comp_name, &arity) == 0 ||
			strcmp (comp_name, ".") != 0 || arity != 2)
		{
			fprintf (stderr, "SP: %d Not a valid kqml list structure\n", msg_no);
			kqml_free_kmsg_list (reply_list);
			*reply_type = KQML_REPLY_ERROR;
			return NULL;
		}
		SP_get_arg (1, SP_reply_list, &SP_kmsg);
		SP_get_arg (2, SP_reply_list, &SP_reply_list);

		/* Where Head is a kmsg */
		kmsg = list_to_kmsg (SP_kmsg);
		if (kmsg == NULL)
		{
			fprintf (stderr, "SP: %d kqml structure conversion error\n", msg_no);
			kqml_free_kmsg_list (reply_list);
			*reply_type = KQML_REPLY_ERROR;
			return NULL;
		}
		else
		{
			if (kqml_build_kmsg_list (&reply_list, kmsg) == KQML_HANDLER_ERROR)
			{
				fprintf (stderr, "SP: %d kqml list construction error\n", msg_no);
				kqml_free_kmsg_list (reply_list);
				*reply_type = KQML_REPLY_ERROR;
				return NULL;
			}
		}
	}

	return reply_list;
}

static void 
string_to_list (int len, char *string, SP_term *SP_list)
{
	int index;
	SP_term SP_char, SP_temp1, SP_temp2;

	SP_show_term (&SP_temp1);
	SP_show_term (&SP_temp2);
	SP_show_term (&SP_char);

	SP_put_string(&SP_temp1, "[]");

	for (index = len - 1; index >= 0; index--)
	{
		SP_put_integer(&SP_char, string[index]);

		SP_put_compound(&SP_temp2, ".", 2);
		SP_put_arg(1, SP_temp2, SP_char);
		SP_put_arg(2, SP_temp2, SP_temp1);

		SP_temp1 = SP_temp2;
	}

	*SP_list = SP_temp1;

	SP_hide_term (&SP_temp1);
	SP_hide_term (&SP_temp2);
	SP_hide_term (&SP_char);
}

static char *
list_to_string (int len, SP_term SP_list)
{
	char *string, *comp_name;
	int index, arity;
	long temp;
	SP_term SP_char;

	string = malloc (len+1);
	if (string == NULL)
	{
		fprintf (stderr, "SP: Not enough memory for string\n");
		return NULL;
	}

	for (index = 0; index < len; index++)
	{
		if (! SP_is_compound (SP_list) ||
			SP_get_compound (SP_list, &comp_name, &arity) == 0 ||
			strcmp (comp_name, ".") != 0 || arity != 2)
		{
			fprintf (stderr, "SP: Not a valid string structure\n");
			free (string);
			return NULL;
		}
		SP_get_arg (1, SP_list, &SP_char);
		SP_get_arg (2, SP_list, &SP_list);

		if (SP_get_integer (SP_char, &temp) == 0 || temp >= 256)
		{
			fprintf (stderr, "SP: Not a valid char\n");
			free (string);
			return NULL;
		}
		string[index] = temp;
	}
	string[len] = '\0';

	return string;
}

static void 
kmsg_to_list (kqml_message *msg, SP_term *SP_list)
{
	int field_no;
	char *field_name, *value;
	int value_len, field_count;
	kqml_expression_type value_type;
	SP_term SP_dummy, SP_param_info, SP_field, SP_value, SP_value_len,
		SP_value_type, SP_temp;

	SP_show_term(&SP_temp);
	SP_show_term(&SP_field);
	SP_show_term(&SP_value_len);
	SP_show_term(&SP_value_type);
	SP_show_term(&SP_dummy);
	SP_show_term(&SP_param_info);

	SP_put_string(&SP_temp, "[]");

	field_count = kqml_get_field_count (msg);
	for (field_no = field_count - 1; field_no >= 0; field_no--)
	{
		if (kqml_get_ith_field (field_no, msg, &field_name, &value,
			&value_len, &value_type) == NULL)
		{
			break;
		}

		SP_put_string(&SP_field, strlwr(field_name));
		string_to_list (value_len, value, &SP_value);
		SP_put_integer(&SP_value_len, value_len);
		SP_put_integer(&SP_value_type, value_type);

		SP_put_compound(&SP_dummy, ",", 2);
		SP_put_arg(1, SP_dummy, SP_value_len);
		SP_put_arg(2, SP_dummy, SP_value_type);
		SP_param_info = SP_dummy;

		SP_put_compound(&SP_dummy, ",", 2);
		SP_put_arg(1, SP_dummy, SP_value);
		SP_put_arg(2, SP_dummy, SP_param_info);
		SP_param_info = SP_dummy;

		SP_put_compound(&SP_dummy, ",", 2);
		SP_put_arg(1, SP_dummy, SP_field);
		SP_put_arg(2, SP_dummy, SP_param_info);
		SP_param_info = SP_dummy;

		SP_put_compound(&SP_dummy, ".", 2);
		SP_put_arg(1, SP_dummy, SP_param_info);
		SP_put_arg(2, SP_dummy, SP_temp);

		SP_temp = SP_dummy;
	}
	
	field_name = kqml_get_performative (msg);
	SP_put_string(&SP_field, strlwr(field_name));

	SP_put_compound(&SP_dummy, ".", 2);
	SP_put_arg(1, SP_dummy, SP_field);
	SP_put_arg(2, SP_dummy, SP_temp);

	*SP_list = SP_dummy;

	SP_hide_term(&SP_temp);
	SP_hide_term(&SP_field);
	SP_hide_term(&SP_value_len);
	SP_hide_term(&SP_value_type);
	SP_hide_term(&SP_dummy);
	SP_hide_term(&SP_param_info);

	return ;
}

static kqml_message *
list_to_kmsg (SP_term SP_list)
{
	int field_no, arity;
	char *comp_name, *field_name, *value;
	long value_len;
	SP_term SP_param_info;
	SP_term SP_field, SP_value_list, SP_value_len, SP_value_type;
	kqml_message *msg;

	/* Check if the prolog list is empty. If it is empty, it will
		be an atom and the name of the atom will be []
	*/
	if (SP_is_atom (SP_list))
	{
		if (SP_get_string (SP_list, &comp_name) == 0 ||
			(strcmp (comp_name, "[]") != 0 && strcmp (comp_name, "nil") != 0))
		{
			fprintf (stderr, "SP: Not a valid kqml structure ?\n");
			return NULL;
		}
		else
		{
			return NULL;
		}
	}

	/* If it is not empty, it should be a compound item of type
		.(performative, rest)
	*/
	if (! SP_is_compound (SP_list) ||
		SP_get_compound (SP_list, &comp_name, &arity) == 0 ||
		strcmp (comp_name, ".") != 0 || arity != 2)
	{
		fprintf (stderr, "SP: Not a valid kqml structure\n");
		return NULL;
	}
	SP_get_arg (1, SP_list, &SP_field);
	SP_get_arg (2, SP_list, &SP_list);

	/* Get the field name */
	if (SP_get_string (SP_field, &field_name) == 0)
	{
		fprintf (stderr, "SP: Not a valid performative\n");
		return NULL;
	}

	msg = kqml_build_msg (field_name, NULL, NULL, NULL, NULL);

	field_no = 0;
	while (++field_no, TRUE)
	{
		/* If this item is the empty list, then we have reached the end 
			of the list */
		if (SP_is_atom (SP_list))
		{
			if (SP_get_string (SP_list, &comp_name) == 0 ||
				strcmp (comp_name, "[]") != 0)
			{
				fprintf (stderr, "SP: %d Not a valid kqml structure ?\n", field_no);
				kqml_free_msg (msg);
				return NULL;
			}
			else
			{
				break;
			}
		}

		/* If this item is not the empty list, it should be
			.(Head, Rest)
		*/
		if (! SP_is_compound (SP_list) ||
			SP_get_compound (SP_list, &comp_name, &arity) == 0 ||
			strcmp (comp_name, ".") != 0 || arity != 2)
		{
			fprintf (stderr, "SP: %d Not a valid kqml structure\n", field_no);
			kqml_free_msg (msg);
			return NULL;
		}
		SP_get_arg (1, SP_list, &SP_param_info);
		SP_get_arg (2, SP_list, &SP_list);

		/* Where Head is a compound structure
			,(Parameter, Rest)
		*/
		if (! SP_is_compound (SP_param_info) ||
			SP_get_compound (SP_param_info, &comp_name, &arity) == 0 ||
			strcmp (comp_name, ",") != 0 || arity != 2)
		{
			 fprintf (stderr, "SP: %d Not a valid kqml parameter structure\n", 
				field_no);
			kqml_free_msg (msg);
			return NULL;
		}
		SP_get_arg (1, SP_param_info, &SP_field);
		SP_get_arg (2, SP_param_info, &SP_param_info);

		/* Where Rest is a compound structure
			,(Value, Rest)
		*/
		if (! SP_is_compound (SP_param_info) ||
			SP_get_compound (SP_param_info, &comp_name, &arity) == 0 ||
			strcmp (comp_name, ",") != 0 || arity != 2)
		{
			fprintf (stderr, "SP: %d Not a valid kqml parameter structure\n", 
				field_no);
			kqml_free_msg (msg);
			return NULL;
		}
		SP_get_arg (1, SP_param_info, &SP_value_list);
		SP_get_arg (2, SP_param_info, &SP_param_info);

		/* Where Rest is a compound strucutre
			,(ValueLen, Rest)
		*/
		if (! SP_is_compound (SP_param_info) ||
			SP_get_compound (SP_param_info, &comp_name, &arity) == 0 ||
			strcmp (comp_name, ",") != 0 || arity != 2)
		{
			fprintf (stderr, "SP: %d Not a valid kqml parameter structure\n", 
				field_no);
			kqml_free_msg (msg);
			return NULL;
		}
		SP_get_arg (1, SP_param_info, &SP_value_len);
		SP_get_arg (2, SP_param_info, &SP_value_type);

		/* Ignore the ValueType field */

		if (SP_get_string (SP_field, &field_name) == 0)
		{
			fprintf (stderr, "SP: %d Not a valid parameter name\n", field_no);
			kqml_free_msg (msg);
			return NULL;
		}

		if (SP_get_integer (SP_value_len, &value_len) == 0)
		{
			fprintf (stderr, "SP: %d Not a valid value length\n", field_no);
			kqml_free_msg (msg);
			return NULL;
		}

		value = list_to_string (value_len, SP_value_list);
		if (value == NULL)
		{
			fprintf (stderr, "SP: %d String Conversion Error\n", field_no);
			kqml_free_msg (msg);
			return NULL;
		}

		if (kqml_put_field (field_name, msg, value, value_len) <= -1)
		{
			kqml_free_msg (msg);
			free (value);
			return NULL;
		}
		free (value);
	}

	return msg;
}

/* Function invoked by the Operating System to handle SIGIO */
static void 
NewSigHandler(int sigNo)
{
	_kqml_dbg_msg(KQML_DBG_VERBOSE)("C_KRIL: Prolog signal handler invoked");

	/* Tell prolog that SpecialHandler should be called, when prolog
		is ready */
	SP_event(SpecialHandler, (void *)sigNo);

	/* Reinstall the signal handler */
	SP_reinstall_signal(sigNo, NewSigHandler);

	_kqml_dbg_msg(KQML_DBG_VERBOSE)("C_KRIL: Passing control back to PROLOG");

	/* Tell prolog that we are done */
	SP_continue();
}

/* Function invoked by prolog to handle SIGIO */
static int
SpecialHandler(int sigNo)
{
	_kqml_dbg_msg(KQML_DBG_VERBOSE)("C_KRIL: KQML signal handler invoked");
	/* Call the actual SIGIO handler (from KQML library) */
	OldSigHandler(sigNo);
	_kqml_dbg_msg(KQML_DBG_VERBOSE)("C_KRIL: KQML signal handler completed");

	/* Return 1, so that prolog can continue its execution */
	/* 0 makes prolog to backtrack and -1 raises an exception */
	return 1;
}

static char *
strlwr (char *string)
{
	char *buffer = string;

	while (*buffer)
	{
		*buffer = tolower(*buffer);
		buffer++;
	}

	return string;
}
