// ******************** type derivation code ****************************
// $Id: TKqmlDeriv.C,v 1.7 1997/06/04 16:35:34 ian Exp $

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <tcl.h>
#include <tk.h>
#include <limits.h>
#include <kqml.h>
#include <TKqml.h>
#include <TKqmlDeriv.h>

int derive_nul(Tcl_Interp *interp, char *s, TTYPE *r, void **nv) {
  return(tkqml_err(TKQML_DERIV_NUL_ERROR)); // who are you trying to kid?
}

int derive_int(Tcl_Interp *interp, char *s, TTYPE *r, void **nv) {
  int rval;

  if ((rval=derive_pint(interp, s, r, nv)) != TCL_OK)
    return(tkqml_err(TKQML_DERIV_INT_ERROR));
  else { // depends on pint allowing values in place of references
    *r=DESTROY; // erase copy instructions from p_int
    //*r = *r&(!DESTROY);
    return(TCL_OK);
  }
}

// pint does not allow actual pointers, but implements call by value-result.
int derive_pint(Tcl_Interp *interp, char *s, TTYPE *r, void **nv) {
  char *v;
  int i;
  int **p = (int **)nv;

  *r=DESTROY;
  if ((v=Tcl_GetVar(interp,s,TCL_GLOBAL_ONLY)) != NULL) { 
    // 1. s is a vbl w/ int string
    if (num_string(v,&i) != TCL_ERROR) {
      *(*p=TALLOC(int))=i; 
      *r=(*r)|KINT|COPY;
    }
    else if (ptr_string(v, nv));                          // is a pointer; done
    else return(tkqml_err(TKQML_DERIV_PINT_ERROR));    // ow, type error - fail
  }
  else if (ptr_string(s, nv));                                 // pointer, done
  else if (num_string(s,&i) != TCL_ERROR)
      *(*p=TALLOC(int))=i; // sneak value through anyway
  else return(tkqml_err(TKQML_DERIV_PINT_ERROR)); 
  return(TCL_OK);
}

int num_string(char *s, int* num) {
  long int lnum;
  if (Tcl_ExprLong (TInterp, s, &lnum) == TCL_ERROR) return (TCL_ERROR);
  if ((lnum >= (LONG_MIN)) && (lnum <= LONG_MAX)) {
      *num = (int)lnum;
      return (TCL_OK);
  }
  else return (TCL_ERROR);
}

int derive_char(Tcl_Interp *interp, char *s, TTYPE *r, void **nv) {
  char *v;

  *r=DESTROY;                                                     /* default */
  if ((v=Tcl_GetVar(interp,s,TCL_GLOBAL_ONLY)) != NULL) {     // using variable
    if (strlen(v)==1) *((char*)*nv=(void*)TALLOC(char))=v[0];
    else return(tkqml_err(TKQML_DERIV_CHAR_ERROR)); // bad char
  }
  else if (strlen(s)==1) *((char*)*nv=(void*)TALLOC(char))=s[0];
  else return(tkqml_err(TKQML_DERIV_CHAR_ERROR));
  return(TCL_OK);
}

// this case spec. designed to avoid interconversion of large content objects
// NOTE: incoming character arrays ALWAYS sized to fit strings; copy
int derive_pchar(Tcl_Interp *interp, char *s, TTYPE *r, void **nv) { 
  char *v;
  
  *r=NO_ACTION;                                                   /* default */
  if ((v=Tcl_GetVar(interp,s,TCL_GLOBAL_ONLY)) != NULL) {       /* variable? */
    if (ptr_string(v,nv));                                  /* pointer; done */
    else {                             /* this is a variable naming a string */
      *r=DESTROY|KSTR|COPY;
      *nv = (void*)strdup(v);
    }
  }
  else if (ptr_string(s,nv));                               /* pointer; done */
  else {
    *nv = (void*)strdup(s);
    *r=DESTROY;
  }
  return(TCL_OK);  /* this call can't fail - everything is at least a string */
}

int derive_pobj(Tcl_Interp *interp, char *s, TTYPE *r, void **nv) {
  char *v;
  
  *r=NO_ACTION;
  if ((v=Tcl_GetVar(interp,s,TCL_GLOBAL_ONLY))) {
    if (!ptr_string(v,nv)) return(tkqml_err(TKQML_DERIV_POBJ_ERROR));
  }
  else if (!ptr_string(s,nv)) { // special case?
    if (strcmp(s,"stdin")==0) *nv=(void *)(stdin);
    else if (strcmp(s,"stdout")==0) *nv=(void *)stdout;
    else if (strcmp(s,"stderr")==0) *nv=(void *)stderr;
    else if (strcmp(s,"NULL")==0) *nv=(void *)NULL;
    else return(tkqml_err(TKQML_DERIV_POBJ_ERROR)); // fail
  }
#ifdef DEBUG
  printf("POBJ returning value %p (stdout=%p)\n",*nv,stdout);
#endif
  return(TCL_OK);
}

int derive_ppobj(Tcl_Interp *interp, char *s, TTYPE *r, void **nv) {
  char *v;
  
  *r=NO_ACTION;
  if ((v=Tcl_GetVar(interp,s,TCL_GLOBAL_ONLY))) {
    *r = COPY | KPPTR; /* | DESTROY */
    if (!ptr_string(v,nv)) return(tkqml_err(TKQML_DERIV_POBJ_ERROR));
    if (strcmp(v, "NULL") == 0) {
      *nv = (void*)malloc(sizeof(void*));
      /* *nv = (void*) TALLOC(void*); */
      /* **nv = NULL; */
    }
  }
  else if (strcmp(s, "NULL") == 0)
    *nv = (void*)NULL;
    // return(tkqml_err(TKQML_DERIV_PPOBJ_ERROR)); // fail
  else if (!ptr_string(s,nv)) {
    return(tkqml_err(TKQML_DERIV_PPOBJ_ERROR)); // fail
  }
#ifdef DEBUG
  printf("PPOBJ returning value %p (stdout=%p)\n",*nv,stdout);
#endif
  return(TCL_OK);
}

// ****************************************************************************
// enumerated type derivation below this line
// ****************************************************************************

int derive_ket(Tcl_Interp *interp, char *s, TTYPE *r, void **nv) {
  kqml_expression_type ket;
  char *v;

  *r=DESTROY;
  if ((v=Tcl_GetVar(interp,s,TCL_GLOBAL_ONLY)) != NULL) {       /* variable? */
    if (ket_enum_lookup(s,&ket)) {                      /* hold name of ket? */
      *nv = (void*)malloc(sizeof(kqml_expression_type));
      *((kqml_expression_type*)*nv) = ket;
    }
    else return(tkqml_err(TKQML_DERIV_KET_ERROR)); // fail
  }
  else {
    if (ket_enum_lookup(v,&ket)) {                      /* hold name of ket? */
      *nv = (void*)malloc(sizeof(kqml_expression_type));
      *((kqml_expression_type*)*nv) = ket;
    }
    else return(tkqml_err(TKQML_DERIV_KET_ERROR)); // fail
  }
  return(TCL_OK);
}

int derive_pket(Tcl_Interp *interp, char *s, TTYPE *r, void **nv) {
  kqml_expression_type ket;
  char *v;

  *r=DESTROY|KKET;
  if ((v=Tcl_GetVar(interp,s,TCL_GLOBAL_ONLY)) != NULL) {       /* variable? */
    if (ptr_string(v, nv)) ;                                   /* ptr - done */
    else if (ket_enum_lookup(s,&ket)) {                 /* hold name of ket? */
      *nv = (void*)malloc(sizeof(kqml_expression_type));
      *((kqml_expression_type*)*nv) = ket;
    }
    *r=*r|COPY;
  }
  else if (ptr_string(s, nv)) ;                                    // ptr -- ok
  else return(tkqml_err(TKQML_DERIV_PKET_ERROR)); // fail
  return(TCL_OK);
}

int derive_krt(Tcl_Interp *interp, char *s, TTYPE *r, void **nv) {
  KQML_REPLY_TYPE krt;
  char *v;

  *r=DESTROY;
  if ((v=Tcl_GetVar(interp,s,TCL_GLOBAL_ONLY)) != NULL) {       /* variable? */
    if (krt_enum_lookup(s,&krt)) {                      /* hold name of krt? */
      *nv = (void*)malloc(sizeof(KQML_REPLY_TYPE));
      *((KQML_REPLY_TYPE*)*nv) = krt;
    }
    else return(tkqml_err(TKQML_DERIV_KRT_ERROR)); // fail
  }
  else {
    if (krt_enum_lookup(v,&krt)) {                      /* hold name of krt? */
      *nv = (void*)malloc(sizeof(KQML_REPLY_TYPE));
      *((KQML_REPLY_TYPE*)*nv) = krt;
    }
    else return(tkqml_err(TKQML_DERIV_KRT_ERROR)); // fail
  }
  return(TCL_OK);
}

int derive_pkrt(Tcl_Interp *interp, char *s, TTYPE *r, void **nv) {
  KQML_REPLY_TYPE krt;
  char *v;
  
  *r=DESTROY|KKRT|COPY;
  if ((v=Tcl_GetVar(interp,s,TCL_GLOBAL_ONLY)) != NULL) {       /* variable? */
    if (ptr_string(v, nv)) ;                               /* pointer - done */
    else if (krt_enum_lookup(s,&krt)) {                 /* hold name of krt? */
      *nv = (void*)malloc(sizeof(KQML_REPLY_TYPE));
      *((KQML_REPLY_TYPE*)*nv) = krt;
    }
  }
  else if (ptr_string(s, nv)) ;
  else return(tkqml_err(TKQML_DERIV_PKRT_ERROR)); // fail
  return (TCL_OK);
}










